@@ -1,10 +1,10 @@
 #include <thread>
 #include <future>
-#include <mutex>
 #include <boost/format.hpp>
 #include <utility>
 #include "version.hh"
 #include "ext/luawrapper/include/LuaContext.hpp"
+#include "lock.hh"
 #include "lua-auth4.hh"
 #include "sstuff.hh"
 #include "minicurl.hh"
@@ -73,7 +73,6 @@ private:
     std::atomic<time_t> lastAccess{0};
   };
 
-  ReadWriteLock d_lock;
 public:
   IsUpOracle()
   {
@@ -158,8 +157,9 @@ private:
       std::vector<std::future<void>> results;
       std::vector<CheckDesc> toDelete;
       {
-        ReadLock lock{&d_lock}; // make sure there's no insertion
-        for (auto& it: d_statuses) {
+        // make sure there's no insertion
+        auto statuses = d_statuses.read_lock();
+        for (auto& it: *statuses) {
           auto& desc = it.first;
           auto& state = it.second;
 
@@ -178,9 +178,9 @@ private:
         future.wait();
       }
       if (!toDelete.empty()) {
-        WriteLock lock{&d_lock};
+        auto statuses = d_statuses.write_lock();
         for (auto& it: toDelete) {
-          d_statuses.erase(it);
+          statuses->erase(it);
         }
       }
       std::this_thread::sleep_until(checkStart + std::chrono::seconds(g_luaHealthChecksInterval));
@@ -188,14 +188,14 @@ private:
   }
 
   typedef map<CheckDesc, std::unique_ptr<CheckState>> statuses_t;
-  statuses_t d_statuses;
+  SharedLockGuarded<statuses_t> d_statuses;
 
   std::unique_ptr<std::thread> d_checkerThread;
 
   void setStatus(const CheckDesc& cd, bool status)
   {
-    ReadLock lock{&d_lock};
-    auto& state = d_statuses[cd];
+    auto statuses = d_statuses.write_lock();
+    auto& state = (*statuses)[cd];
     state->status = status;
     if (state->first) {
       state->first = false;
@@ -229,13 +229,13 @@ private:
 bool IsUpOracle::isUp(const CheckDesc& cd)
 {
   if (!d_checkerThread) {
-    d_checkerThread = std::unique_ptr<std::thread>(new std::thread(&IsUpOracle::checkThread, this));
+    d_checkerThread = std::make_unique<std::thread>([this] { return checkThread(); });
   }
   time_t now = time(nullptr);
   {
-    ReadLock lock{&d_lock};
-    auto iter = d_statuses.find(cd);
-    if (iter != d_statuses.end()) {
+    auto statuses = d_statuses.read_lock();
+    auto iter = statuses->find(cd);
+    if (iter != statuses->end()) {
       iter->second->lastAccess = now;
       return iter->second->status;
     }
@@ -245,10 +245,10 @@ bool IsUpOracle::isUp(const CheckDesc& cd)
     ComboAddress src(cd.opts.at("source"));
   }
   {
-    WriteLock lock{&d_lock};
+    auto statuses = d_statuses.write_lock();
     // Make sure we don't insert new entry twice now we have the lock
-    if (d_statuses.find(cd) == d_statuses.end()) {
-      d_statuses[cd] = std::unique_ptr<CheckState>(new CheckState{now});
+    if (statuses->find(cd) == statuses->end()) {
+      (*statuses)[cd] = std::make_unique<CheckState>(now);
     }
   }
   return false;
@@ -326,7 +326,7 @@ static ComboAddress pickwrandom(const vector<pair<int,ComboAddress> >& wips)
   vector<pair<int, ComboAddress> > pick;
   for(auto& i : wips) {
     sum += i.first;
-    pick.push_back({sum, i.second});
+    pick.emplace_back(sum, i.second);
   }
   int r = dns_random(sum);
   auto p = upper_bound(pick.begin(), pick.end(), r, [](int rarg, const decltype(pick)::value_type& a) { return rarg < a.first; });
@@ -443,14 +443,16 @@ static ComboAddress pickclosest(const ComboAddress& bestwho, const vector<ComboA
 
 static std::vector<DNSZoneRecord> lookup(const DNSName& name, uint16_t qtype, int zoneid)
 {
-  static UeberBackend ub;
-  static std::mutex mut;
-  std::lock_guard<std::mutex> lock(mut);
-  ub.lookup(QType(qtype), name, zoneid);
+  static LockGuarded<UeberBackend> s_ub;
+
   DNSZoneRecord dr;
   vector<DNSZoneRecord> ret;
-  while(ub.get(dr)) {
-    ret.push_back(dr);
+  {
+    auto ub = s_ub.lock();
+    ub->lookup(QType(qtype), name, zoneid);
+    while (ub->get(dr)) {
+      ret.push_back(dr);
+    }
   }
   return ret;
 }
@@ -970,7 +972,7 @@ std::vector<shared_ptr<DNSRecordContent>> luaSynth(const std::string& code, cons
 
   LuaContext& lua = *s_LUA->getLua();
 
-  s_lua_record_ctx = std::unique_ptr<lua_record_ctx_t>(new lua_record_ctx_t());
+  s_lua_record_ctx = std::make_unique<lua_record_ctx_t>();
   s_lua_record_ctx->qname = query;
   s_lua_record_ctx->zone = zone;
   s_lua_record_ctx->zoneid = zoneid;
@@ -978,7 +980,7 @@ std::vector<shared_ptr<DNSRecordContent>> luaSynth(const std::string& code, cons
   lua.writeVariable("qname", query);
   lua.writeVariable("zone", zone);
   lua.writeVariable("zoneid", zoneid);
-  lua.writeVariable("who", dnsp.getRemote());
+  lua.writeVariable("who", dnsp.getInnerRemote());
   lua.writeVariable("dh", (dnsheader*)&dnsp.d);
   lua.writeVariable("dnssecOK", dnsp.d_dnssecOk);
   lua.writeVariable("tcp", dnsp.d_tcp);
@@ -989,7 +991,7 @@ std::vector<shared_ptr<DNSRecordContent>> luaSynth(const std::string& code, cons
   }
   else {
     lua.writeVariable("ecswho", nullptr);
-    s_lua_record_ctx->bestwho = dnsp.getRemote();
+    s_lua_record_ctx->bestwho = dnsp.getInnerRemote();
   }
   lua.writeVariable("bestwho", s_lua_record_ctx->bestwho);
 
